from typing import Self, override

from luma.core.super import Estimator, Evaluator, Optimizer, Supervised
from luma.interface.typing import Matrix, Tensor, Vector
from luma.interface.util import InitUtil
from luma.metric.classification import Accuracy

from luma.neural import loss
from luma.neural.base import Loss, NeuralModel
from luma.neural.block import ConvBlock2D, DenseBlock
from luma.neural.layer import Activation, Dense, Flatten, Sequential


__all__ = ("_LeNet_1", "_LeNet_4", "_LeNet_5")


class _LeNet_1(Estimator, Supervised, NeuralModel):
    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.Tanh,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 10,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0.0,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int = None,
        deep_verbose: bool = False,
    ) -> None:
        self.activation = activation
        self.optimizer = optimizer
        self.loss = loss
        self.initializer = initializer
        self.out_features = out_features
        self.lambda_ = lambda_
        self.shuffle = shuffle
        self.random_state = random_state
        self._fitted = False

        super().__init__(
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            early_stopping,
            patience,
            deep_verbose,
        )
        super().__init_model__()
        self.model = Sequential()
        self.optimizer.set_params(learning_rate=self.learning_rate)
        self.model.set_optimizer(optimizer=self.optimizer)

        self.feature_sizes_ = [
            [1, 4, 8],
            [8 * 4 * 4, self.out_features],
        ]
        self.feature_shapes_ = [
            [(1, 4), (4, 8)],
            [(8 * 4 * 4, self.out_features)],
        ]

        self.set_param_ranges(
            {
                "out_features": ("0<,+inf", int),
                "batch_size": ("0<,+inf", int),
                "n_epochs": ("0<,+inf", int),
                "learning_rate": ("0<,+inf", None),
                "valid_size": ("0<,<1", None),
                "dropout_rate": ("0,1", None),
                "lambda_": ("0,+inf", None),
                "patience": (f"0<,+inf", int),
            }
        )
        self.check_param_ranges()
        self._build_model()

    def _build_model(self) -> None:
        self.model += ConvBlock2D(
            1,
            4,
            filter_size=5,
            stride=1,
            activation=self.activation,
            initializer=self.initializer,
            padding="valid",
            lambda_=self.lambda_,
            do_batch_norm=False,
            pool_filter_size=2,
            pool_stride=2,
            pool_mode="avg",
            random_state=self.random_state,
        )
        self.model += ConvBlock2D(
            4,
            8,
            filter_size=5,
            stride=1,
            activation=self.activation,
            initializer=self.initializer,
            padding="valid",
            lambda_=self.lambda_,
            do_batch_norm=False,
            pool_filter_size=2,
            pool_stride=2,
            pool_mode="avg",
            random_state=self.random_state,
        )

        self.model += Flatten()
        self.model += Dense(
            8 * 4 * 4,
            self.out_features,
            lambda_=self.lambda_,
            random_state=self.random_state,
        )

    @Tensor.force_dim(4)
    def fit(self, X: Tensor, y: Matrix) -> Self:
        return super(_LeNet_1, self).fit_nn(X, y)

    @override
    @Tensor.force_dim(4)
    def predict(self, X: Tensor, argmax: bool = True) -> Matrix | Vector:
        return super(_LeNet_1, self).predict_nn(X, argmax)

    @override
    @Tensor.force_dim(4)
    def score(
        self,
        X: Tensor,
        y: Matrix,
        metric: Evaluator = Accuracy,
        argmax: bool = True,
    ) -> float:
        return super(_LeNet_1, self).score_nn(X, y, metric, argmax)


class _LeNet_4(Estimator, Supervised, NeuralModel):
    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.Tanh,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 10,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0.0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int = None,
        deep_verbose: bool = False,
    ) -> None:
        self.activation = activation
        self.optimizer = optimizer
        self.loss = loss
        self.initializer = initializer
        self.out_features = out_features
        self.lambda_ = lambda_
        self.dropout_rate = dropout_rate
        self.shuffle = shuffle
        self.random_state = random_state
        self._fitted = False

        super().__init__(
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            early_stopping,
            patience,
            deep_verbose,
        )
        super().__init_model__()
        self.model = Sequential()
        self.optimizer.set_params(learning_rate=self.learning_rate)
        self.model.set_optimizer(optimizer=self.optimizer)

        self.feature_sizes_ = [
            [1, 4, 16],
            [16 * 5 * 5, 120, self.out_features],
        ]
        self.feature_shapes_ = [
            [(1, 4), (4, 16)],
            [(16 * 5 * 5, 120), (120, self.out_features)],
        ]

        self.set_param_ranges(
            {
                "out_features": ("0<,+inf", int),
                "batch_size": ("0<,+inf", int),
                "n_epochs": ("0<,+inf", int),
                "learning_rate": ("0<,+inf", None),
                "valid_size": ("0<,<1", None),
                "dropout_rate": ("0,1", None),
                "lambda_": ("0,+inf", None),
                "patience": (f"0<,+inf", int),
            }
        )
        self.check_param_ranges()
        self._build_model()

    def _build_model(self) -> None:
        self.model += ConvBlock2D(
            1,
            4,
            filter_size=5,
            stride=1,
            activation=self.activation,
            initializer=self.initializer,
            padding="valid",
            lambda_=self.lambda_,
            do_batch_norm=False,
            pool_filter_size=2,
            pool_stride=2,
            pool_mode="avg",
            random_state=self.random_state,
        )
        self.model += ConvBlock2D(
            4,
            16,
            filter_size=5,
            stride=1,
            activation=self.activation,
            initializer=self.initializer,
            padding="valid",
            lambda_=self.lambda_,
            do_batch_norm=False,
            pool_filter_size=2,
            pool_stride=2,
            pool_mode="avg",
            random_state=self.random_state,
        )

        self.model += Flatten()
        self.model += DenseBlock(
            16 * 5 * 5,
            120,
            activation=self.activation,
            lambda_=self.lambda_,
            do_batch_norm=False,
            do_dropout=False,
            random_state=self.random_state,
        )
        self.model += Dense(
            120,
            self.out_features,
            lambda_=self.lambda_,
            random_state=self.random_state,
        )

    @Tensor.force_dim(4)
    def fit(self, X: Tensor, y: Matrix) -> Self:
        return super(_LeNet_4, self).fit_nn(X, y)

    @override
    @Tensor.force_dim(4)
    def predict(self, X: Tensor, argmax: bool = True) -> Matrix | Vector:
        return super(_LeNet_4, self).predict_nn(X, argmax)

    @override
    @Tensor.force_dim(4)
    def score(
        self,
        X: Tensor,
        y: Matrix,
        metric: Evaluator = Accuracy,
        argmax: bool = True,
    ) -> float:
        return super(_LeNet_4, self).score_nn(X, y, metric, argmax)


class _LeNet_5(Estimator, Supervised, NeuralModel):
    def __init__(
        self,
        optimizer: Optimizer,
        activation: Activation.FuncType = Activation.Tanh,
        loss: Loss = loss.CrossEntropy(),
        initializer: InitUtil.InitStr = None,
        out_features: int = 10,
        batch_size: int = 100,
        n_epochs: int = 100,
        learning_rate: float = 0.01,
        valid_size: float = 0.1,
        lambda_: float = 0.0,
        dropout_rate: float = 0.5,
        early_stopping: bool = False,
        patience: int = 10,
        shuffle: bool = True,
        random_state: int = None,
        deep_verbose: bool = False,
    ) -> None:
        self.activation = activation
        self.optimizer = optimizer
        self.loss = loss
        self.initializer = initializer
        self.out_features = out_features
        self.lambda_ = lambda_
        self.dropout_rate = dropout_rate
        self.shuffle = shuffle
        self.random_state = random_state
        self._fitted = False

        super().__init__(
            batch_size,
            n_epochs,
            learning_rate,
            valid_size,
            early_stopping,
            patience,
            deep_verbose,
        )
        super().__init_model__()
        self.model = Sequential()
        self.optimizer.set_params(learning_rate=self.learning_rate)
        self.model.set_optimizer(optimizer=self.optimizer)

        self.feature_sizes_ = [
            [1, 6, 16],
            [16 * 5 * 5, 120, 84, self.out_features],
        ]
        self.feature_shapes_ = [
            [(1, 6), (6, 16)],
            [(16 * 5 * 5, 120), (120, 84), (84, self.out_features)],
        ]

        self.set_param_ranges(
            {
                "out_features": ("0<,+inf", int),
                "batch_size": ("0<,+inf", int),
                "n_epochs": ("0<,+inf", int),
                "learning_rate": ("0<,+inf", None),
                "valid_size": ("0<,<1", None),
                "dropout_rate": ("0,1", None),
                "lambda_": ("0,+inf", None),
                "patience": (f"0<,+inf", int),
            }
        )
        self.check_param_ranges()
        self._build_model()

    def _build_model(self) -> None:
        self.model += ConvBlock2D(
            1,
            6,
            filter_size=5,
            stride=1,
            activation=self.activation,
            initializer=self.initializer,
            padding="valid",
            lambda_=self.lambda_,
            do_batch_norm=False,
            pool_filter_size=2,
            pool_stride=2,
            pool_mode="avg",
            random_state=self.random_state,
        )
        self.model += ConvBlock2D(
            6,
            16,
            filter_size=5,
            stride=1,
            activation=self.activation,
            initializer=self.initializer,
            padding="valid",
            lambda_=self.lambda_,
            do_batch_norm=False,
            pool_filter_size=2,
            pool_stride=2,
            pool_mode="avg",
            random_state=self.random_state,
        )

        self.model += Flatten()
        self.model += DenseBlock(
            16 * 5 * 5,
            120,
            activation=self.activation,
            lambda_=self.lambda_,
            do_batch_norm=False,
            do_dropout=False,
            random_state=self.random_state,
        )
        self.model += DenseBlock(
            120,
            84,
            activation=self.activation,
            lambda_=self.lambda_,
            do_batch_norm=False,
            do_dropout=False,
            random_state=self.random_state,
        )
        self.model += Dense(
            84,
            self.out_features,
            lambda_=self.lambda_,
            random_state=self.random_state,
        )

    @Tensor.force_dim(4)
    def fit(self, X: Tensor, y: Matrix) -> Self:
        return super(_LeNet_5, self).fit_nn(X, y)

    @override
    @Tensor.force_dim(4)
    def predict(self, X: Tensor, argmax: bool = True) -> Matrix | Vector:
        return super(_LeNet_5, self).predict_nn(X, argmax)

    @override
    @Tensor.force_dim(4)
    def score(
        self,
        X: Tensor,
        y: Matrix,
        metric: Evaluator = Accuracy,
        argmax: bool = True,
    ) -> float:
        return super(_LeNet_5, self).score_nn(X, y, metric, argmax)



from typing import Self, Tuple
import numpy as np

from luma.core.super import Estimator, Evaluator, Supervised
from luma.interface.typing import Matrix, Vector
from luma.interface.exception import NotFittedError
from luma.metric.regression import MeanSquaredError
from luma.regressor.linear import LinearRegressor


__all__ = ("RANSAC", "MLESAC")


class RANSAC(Estimator, Supervised):
    """
    RANSAC (Random Sample Consensus) is an iterative algorithm used for robust
    estimation of parameters from a set of data containing outliers. It works by
    randomly selecting a subset of the original data and fitting a model to this
    subset. The algorithm then tests all other data points against this model,
    classifying them as inliers or outliers based on a predefined threshold.
    This process is repeated multiple times, each time selecting a different
    subset. The best model is chosen based on the maximum number of inliers
    it includes.

    Parameters
    ----------
    `estimator` : Estimator, default=LinearRegressor()
        Regression estimator
    `min_points` : int or float, default=2
        Mininum sample size for each random sample
    `max_iter` : int, default=1000
        Maximum iteration
    `min_inliers` : int or float, default=0.5
        Minimum number of inliers for a model to be considered valid
    `threshold` : float, optional, default=None
        Maximum distance a point can have to be considered an inlier
    `random_state` : int, optional, default=None
        Seed for random sampling the data

    Properties
    ----------
    Get best estimator:
    ```py
        @property
        def best_estimator(self) -> Estimator
    ```
    """

    def __init__(
        self,
        estimator: Estimator = LinearRegressor(),
        min_points: int | float = 2,
        max_iter: int = 1000,
        min_inliers: int | float = 0.5,
        threshold: float | None = None,
        random_state: int | None = None,
        verbose: bool = False,
    ) -> None:
        self.estimator = estimator
        self.min_points = min_points
        self.max_iter = max_iter
        self.min_inliers = min_inliers
        self.threshold = threshold
        self.random_state = random_state
        self.verbose = verbose
        self.inliers_ = None
        self._fitted = False

        self.set_param_ranges(
            {
                "min_points": ("0,+inf", int),
                "max_iter": ("0<,+inf", int),
                "min_inliers": ("0,+inf", None),
                "threshold": ("0,+inf", None),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> Self:
        m, _ = X.shape
        best_err = np.inf
        best_indices = None

        np.random.seed(self.random_state)
        Xy = np.hstack((X, y.reshape(-1, 1)))

        if self.min_points < 1:
            self.min_points = np.ceil(m * self.min_points)
        if self.min_inliers < 1:
            self.min_inliers = np.ceil(m * self.min_inliers)
        if self.threshold is None:
            self.threshold = self._mean_absdev(y)

        for i in range(self.max_iter):
            hypo_indices, test_indices = self._random_partition(X)
            hypo_inliers = Xy[hypo_indices]
            test_points = Xy[test_indices]

            self.estimator.fit(hypo_inliers[:, :-1], hypo_inliers[:, -1])
            test_err = self._error(test_points[:, :-1], test_points[:, -1])

            true_indices = test_indices[test_err < self.threshold]
            true_inliers = Xy[true_indices]

            if len(true_inliers) > self.min_inliers:
                Xy_better = np.concatenate((hypo_inliers, true_inliers))

                self.estimator.fit(Xy_better[:, :-1], Xy_better[:, -1])
                better_err = self._error(Xy_better[:, :-1], Xy_better[:, -1])
                this_err = np.mean(better_err)

                if this_err < best_err:
                    best_err = this_err
                    best_indices = np.concatenate((hypo_indices, true_indices))

            if self.verbose and i % 100 == 0:
                print(
                    f"[RANSAC] Iteration {i}/{self.max_iter}",
                    f"with current best-error: {best_err}",
                )

        self.inliers_ = Xy[best_indices, :-1]
        self._fitted = True
        return self

    def _random_partition(self, X: Matrix) -> Tuple[Vector, Vector]:
        all_indices = np.arange(X.shape[0])
        np.random.shuffle(all_indices)
        indices_l = all_indices[: self.min_points]
        indices_r = all_indices[self.min_points :]

        return indices_l, indices_r

    def _error(self, X: Matrix, y: Vector) -> Vector:
        y_pred = self.estimator.predict(X)
        return np.abs(y - y_pred)

    def _mean_absdev(self, y: Vector) -> float:
        y_median = np.median(y)
        abs_dev = np.abs(y - y_median)
        mad = np.median(abs_dev)
        return mad

    @property
    def best_estimator(self) -> Estimator:
        if not self._fitted:
            raise NotFittedError(self)
        return self.estimator

    def predict(self, X: Matrix) -> Vector:
        best = self.best_estimator
        return best.predict(X)

    def score(
        self, X: Matrix, y: Matrix, metric: Evaluator = MeanSquaredError
    ) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)


class MLESAC(Estimator, Supervised):
    """
    MLESAC (Maximum Likelihood Estimation SAmple Consensus) is an iterative algorithm
    used for robust estimation of parameters from a set of data containing outliers.
    Unlike RANSAC which focuses on maximizing the number of inliers, MLESAC seeks
    to maximize the likelihood of the model given the data, considering both the
    number of inliers and the quality of the fit.

    Parameters
    ----------
    `estimator` : Estimator, default=LinearRegressor()
        Regression estimator
    `min_points` : int or float, default=2
        Minimum sample size for each random sample
    `max_iter` : int, default=1000
        Maximum number of iterations
    `min_inliers` : int or float, default=0.5
        Minimum number of inliers for a model to be considered valid
    `threshold` : float, optional, default=None
        Maximum distance a point can have to be considered an inlier
    `random_state` : int, optional, default=None
        Seed for random sampling the data

    Properties
    ----------
    Get best estimator:
    ```py
        @property
        def best_estimator(self) -> Estimator
    ```

    Notes
    -----
    * `MLESAC` assumes that inliers follow a multivariate Gaussian model

    """

    def __init__(
        self,
        estimator: Estimator = LinearRegressor(),
        min_points: int | float = 2,
        max_iter: int = 1000,
        min_inliers: int | float = 0.5,
        threshold: float | None = None,
        random_state: int | None = None,
        verbose: bool = False,
    ) -> None:
        self.estimator = estimator
        self.min_points = min_points
        self.max_iter = max_iter
        self.min_inliers = min_inliers
        self.threshold = threshold
        self.random_state = random_state
        self.verbose = verbose
        self.inliers_ = None
        self._fitted = False

        self.set_param_ranges(
            {
                "min_points": ("0,+inf", int),
                "max_iter": ("0<,+inf", int),
                "min_inliers": ("0,+inf", None),
                "threshold": ("0,+inf", None),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> Self:
        m, _ = X.shape
        best_likelihood = -np.inf
        best_indices = None

        np.random.seed(self.random_state)
        Xy = np.hstack((X, y.reshape(-1, 1)))

        if self.min_points < 1:
            self.min_points = np.ceil(m * self.min_points)
        if self.min_inliers < 1:
            self.min_inliers = np.ceil(m * self.min_inliers)
        if self.threshold is None:
            self.threshold = self._mean_absdev(y)

        for i in range(self.max_iter):
            hypo_indices, test_indices = self._random_partition(X)
            hypo_inliers = Xy[hypo_indices]
            test_points = Xy[test_indices]

            self.estimator.fit(hypo_inliers[:, :-1], hypo_inliers[:, -1])
            test_err = self._error(test_points[:, :-1], test_points[:, -1])

            true_indices = test_indices[test_err < self.threshold]
            true_inliers = Xy[true_indices]

            if len(true_inliers) >= self.min_inliers:
                Xy_better = np.concatenate((hypo_inliers, true_inliers))

                self.estimator.fit(Xy_better[:, :-1], Xy_better[:, -1])
                better_err = self._error(Xy_better[:, :-1], Xy_better[:, -1])
                likelihood = self._calculate_likelihood(better_err)

                if likelihood > best_likelihood:
                    best_likelihood = likelihood
                    best_indices = np.concatenate((hypo_indices, true_indices))

            if self.verbose and i % 100 == 0:
                print(
                    f"[MLESAC] Iteration {i}/{self.max_iter}",
                    f"with current best-likelihood: {best_likelihood}",
                )

        self.inliers_ = Xy[best_indices, :-1]
        self._fitted = True
        return self

    def _random_partition(self, X: Matrix) -> Tuple[Vector, Vector]:
        all_indices = np.arange(X.shape[0])
        np.random.shuffle(all_indices)
        indices_l = all_indices[: self.min_points]
        indices_r = all_indices[self.min_points :]

        return indices_l, indices_r

    def _error(self, X: Matrix, y: Vector) -> Vector:
        y_pred = self.estimator.predict(X)
        return np.abs(y - y_pred)

    def _mean_absdev(self, y: Vector) -> float:
        y_median = np.median(y)
        abs_dev = np.abs(y - y_median)
        mad = np.median(abs_dev)
        return mad

    def _calculate_likelihood(self, error: Vector) -> float:
        if error.ndim == 1:
            error = error[:, np.newaxis]
        m, n = error.shape
        cov_matrix = np.cov(error, rowvar=False)

        if n == 1:
            cov_matrix = cov_matrix.reshape(1, 1)
        det_cov_matrix = np.linalg.det(cov_matrix)
        inv_cov_matrix = np.linalg.inv(cov_matrix)

        likelihood = -m * n * np.log(2 * np.pi) / 2
        likelihood -= m * np.log(det_cov_matrix) / 2
        for error in error:
            likelihood -= 0.5 * np.dot(error, np.dot(inv_cov_matrix, error))

        return likelihood

    @property
    def best_estimator(self) -> Estimator:
        if not self._fitted:
            raise NotFittedError(self)
        return self.estimator

    def predict(self, X: Matrix) -> Vector:
        best = self.best_estimator
        return best.predict(X)

    def score(
        self, X: Matrix, y: Matrix, metric: Evaluator = MeanSquaredError
    ) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)



import numpy as np

from luma.interface.typing import Matrix, ClassType
from luma.core.super import Evaluator


__all__ = (
    "MeanAbsoluteError",
    "MeanSquaredError",
    "RootMeanSquaredError",
    "MeanAbsolutePercentageError",
    "RSquaredScore",
    "AdjustedRSquaredScore",
)


@ClassType.non_instantiable()
class MeanAbsoluteError(Evaluator):
    @classmethod
    def score(cls, y_true: Matrix, y_pred: Matrix) -> float:
        return np.mean(np.abs(y_true - y_pred))


@ClassType.non_instantiable()
class MeanSquaredError(Evaluator):
    @classmethod
    def score(cls, y_true: Matrix, y_pred: Matrix) -> float:
        return np.mean((y_true - y_pred) ** 2)


@ClassType.non_instantiable()
class RootMeanSquaredError(Evaluator):
    @classmethod
    def score(cls, y_true: Matrix, y_pred: Matrix) -> float:
        return np.sqrt(np.mean((y_true - y_pred) ** 2))


@ClassType.non_instantiable()
class MeanAbsolutePercentageError(Evaluator):
    @classmethod
    def score(cls, y_true: Matrix, y_pred: Matrix) -> float:
        return np.mean(np.abs((y_true - y_pred) / y_true)) * 100


@ClassType.non_instantiable()
class RSquaredScore(Evaluator):
    @classmethod
    def score(cls, y_true: Matrix, y_pred: Matrix) -> float:
        y_bar = np.mean(y_true)
        sst = np.sum((y_true - y_bar) ** 2)
        ssr = np.sum((y_true - y_pred) ** 2)
        r2 = 1 - (ssr / sst)
        return r2


@ClassType.non_instantiable()
class AdjustedRSquaredScore(Evaluator):
    @classmethod
    def score(cls, y_true: Matrix, y_pred: Matrix, n_predictors: int) -> float:
        m = len(y_true)
        y_bar = np.mean(y_true)
        sst = np.sum((y_true - y_bar) ** 2)
        ssr = np.sum((y_true - y_pred) ** 2)
        r2 = 1 - (ssr / sst)
        r2_adj = 1 - ((1 - r2) * (m - 1) / (m - n_predictors - 1))
        return r2_adj



from typing import Self
import numpy as np

from luma.interface.typing import Matrix, Scalar
from luma.core.super import Transformer
from luma.interface.exception import NotFittedError


__all__ = ("StandardScaler", "MinMaxScaler")


class StandardScaler(Transformer):
    """
    Standard scaling is a data preprocessing technique to transform
    the data so that it has a mean of 0 and a standard deviation of 1.
    """

    def __init__(self) -> None:
        self.mean = None
        self.std = None
        self._fitted = False

    def fit(self, X: Matrix) -> Self:
        self.mean = np.mean(X, axis=0)
        self.std = np.std(X, axis=0, ddof=1)
        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        epsilon = 1e-8
        return (X - self.mean) / (self.std + epsilon)

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform(X)

    def inverse_transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return (X * self.std) + self.mean


class MinMaxScaler(Transformer, Transformer.Feature):
    """
    MinMax scaling (also known as Min-Max normalization)
    to rescale features to a specific range, typically between 0 and 1.
    The purpose of MinMax scaling is to transform the features in a way
    that they fall within a specific interval.

    Parameters
    ----------
    `feature_range` : tuple[Scalar, Scalar], default=(0,1)
        Range to be scaled

    """

    def __init__(self, feature_range: tuple[Scalar, Scalar] = (0, 1)) -> None:
        self.feature_range = feature_range
        self.min = None
        self.max = None
        self._fitted = False

    def fit(self, X: Matrix) -> Self:
        self.min = np.min(X, axis=0)
        self.max = np.max(X, axis=0)
        self._fitted = True
        return self

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        min_val, max_val = self.feature_range
        scaled = (X - self.min) / (self.max - self.min) * (max_val - min_val)
        return scaled + min_val

    def fit_transform(self, X: Matrix) -> Matrix:
        self.fit(X)
        return self.transform(X)

    def inverse_transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        min_val, max_val = self.feature_range
        original = (X - min_val) / (max_val - min_val) * (self.max - self.min)
        return original + self.min



from typing import Tuple, Type
import numpy as np

from luma.interface.typing import ClassType, Tensor
from luma.neural.base import Layer


__all__ = "_Activation"


@ClassType.non_instantiable()
class _Activation:
    type FuncType = Type

    @classmethod
    def _out_shape(cls, in_shape: Tuple[int]) -> Tuple[int]:
        return in_shape

    class Linear(Layer):
        def __init__(self) -> None:
            super().__init__()

        def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
            _ = is_train
            return X

        def backward(self, d_out: Tensor) -> Tensor:
            self.dX = d_out
            return self.dX

        def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
            return _Activation._out_shape(in_shape)

    class ReLU(Layer):
        def __init__(self) -> None:
            super().__init__()

        def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
            _ = is_train
            self.input_ = X
            return np.maximum(0, X)

        def backward(self, d_out: Tensor) -> Tensor:
            self.dX = d_out.copy()
            self.dX[self.input_ <= 0] = 0
            return self.dX

        def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
            return _Activation._out_shape(in_shape)

    class Sigmoid(Layer):
        def __init__(self) -> None:
            super().__init__()

        def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
            _ = is_train
            self.output_ = 1 / (1 + np.exp(-X))
            return self.output_

        def backward(self, d_out: Tensor) -> Tensor:
            self.dX = d_out * self.output_ * (1 - self.output_)
            return self.dX

        def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
            return _Activation._out_shape(in_shape)

    class Tanh(Layer):
        def __init__(self) -> None:
            super().__init__()

        def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
            _ = is_train
            self.output_ = np.tanh(X)
            return self.output_

        def backward(self, d_out: Tensor) -> Tensor:
            self.dX = d_out * (1 - np.square(self.output_))
            return self.dX

        def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
            return _Activation._out_shape(in_shape)

    class LeakyReLU(Layer):
        def __init__(self, alpha: float = 0.01) -> None:
            super().__init__()
            self.alpha = alpha

        def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
            _ = is_train
            self.input_ = X
            return np.where(X > 0, X, X * self.alpha)

        def backward(self, d_out: Tensor) -> Tensor:
            self.dX = d_out * np.where(self.input_ > 0, 1, self.alpha)
            return self.dX

        def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
            return _Activation._out_shape(in_shape)

    class Softmax(Layer):
        def __init__(self) -> None:
            super().__init__()

        def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
            _ = is_train
            e_X = np.exp(X - np.max(X, axis=-1, keepdims=True))
            return e_X / np.sum(e_X, axis=-1, keepdims=True)

        def backward(self, d_out: Tensor) -> Tensor:
            self.dX = np.empty_like(d_out)
            for i, (y, dy) in enumerate(zip(self.output_, d_out)):
                y = y.reshape(-1, 1)
                jacobian_matrix = np.diagflat(y) - np.dot(y, y.T)

                self.dX[i] = np.dot(jacobian_matrix, dy)

            return self.dX

        def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
            return _Activation._out_shape(in_shape)

    class ELU(Layer):
        def __init__(self, alpha: float = 1.0) -> None:
            super().__init__()
            self.alpha = alpha

        def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
            _ = is_train
            self.input_ = X
            self.output_ = np.where(X > 0, X, self.alpha * (np.exp(X) - 1))
            return self.output_

        def backward(self, d_out: Tensor) -> Tensor:
            self.dX = d_out * np.where(
                self.input_ > 0,
                1,
                self.output_ + self.alpha,
            )
            return self.dX

        def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
            return _Activation._out_shape(in_shape)

    class SELU(Layer):
        def __init__(
            self,
            lambda_: float = 1.0507,
            alpha: float = 1.67326,
        ) -> None:
            super().__init__()
            self.lambda_ = lambda_
            self.alpha = alpha

        def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
            _ = is_train
            self.input_ = X
            self.output_ = self.lambda_ * np.where(
                X > 0, X, self.alpha * (np.exp(X) - 1)
            )
            return self.output_

        def backward(self, d_out: Tensor) -> Tensor:
            self.dX = (
                d_out
                * self.lambda_
                * np.where(
                    self.input_ > 0,
                    1,
                    self.alpha * np.exp(self.input_),
                )
            )
            return self.dX

        def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
            return _Activation._out_shape(in_shape)

    class Softplus(Layer):
        def __init__(self) -> None:
            super().__init__()

        def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
            _ = is_train
            self.output_ = np.log1p(np.exp(X))
            return self.output_

        def backward(self, d_out: Tensor) -> Tensor:
            self.dX = d_out * (1 - 1 / (1 + np.exp(self.output_)))
            return self.dX

        def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
            return _Activation._out_shape(in_shape)

    class Swish(Layer):
        def __init__(self, beta: float = 1.0) -> None:
            super().__init__()
            self.beta = beta

        def forward(self, X: Tensor, is_train: bool = False) -> Tensor:
            _ = is_train
            self.input_ = X
            self.sigmoid = 1 / (1 + np.exp(-self.beta * X))

            self.output_ = X * self.sigmoid
            return self.output_

        def backward(self, d_out: Tensor) -> Tensor:
            self.dX = d_out * (
                self.sigmoid
                + self.beta * self.input_ * self.sigmoid * (1 - self.sigmoid)
            )
            return self.dX

        def out_shape(self, in_shape: Tuple[int]) -> Tuple[int]:
            return _Activation._out_shape(in_shape)



from typing import Tuple
import numpy as np

from luma.core.super import Estimator, Evaluator
from luma.interface.typing import Matrix, Vector
from luma.model_selection.fold import FoldType, KFold


__all__ = "CrossValidator"


class CrossValidator(Evaluator):
    """
    Cross-validation is a technique in machine learning for assessing how
    a model generalizes to an independent dataset. It involves dividing
    the data into several parts, training the model on some parts and
    testing it on others. This process is repeated multiple times with
    different partitions, providing a robust estimate of the model's
    performance. It's especially useful in scenarios with limited data,
    ensuring that all available data is effectively utilized for both
    training and validation.

    Parameters
    ----------
    `estimator` : Estimator
        An estimator to validate
    `metric` : Evaluator
        Evaluation metric for validation
    `cv` : int, default=5
        Number of folds in splitting data
    `fold_type` : FoldType, default=KFold
        Fold type
    `shuffle` : bool, default=True
        Whether to shuffle the dataset
    `random_state` : int, optional, default=None
        Seed for random sampling upon splitting data

    Attributes
    ----------
    `train_scores_` : list
        List of training scores for each fold
    `test_scores_` : list
        List of test(validation) scores for each fold

    Methods
    -------
    For getting mean train score and mean test score respectvely:
    ```py
        def score(self, X: Matrix, y: Vector) -> Tuple[float, float]
    ```
    """

    def __init__(
        self,
        estimator: Estimator,
        metric: Evaluator,
        cv: int = 5,
        fold_type: FoldType = KFold,
        shuffle: bool = True,
        random_state: int | None = None,
        verbose: bool = False,
    ) -> None:
        self.estimator = estimator
        self.metric = metric
        self.cv = cv
        self.fold_type = fold_type
        self.shuffle = shuffle
        self.random_state = random_state
        self.verbose = verbose
        self.train_scores_ = []
        self.test_scores_ = []

    def _fit(self, X: Matrix, y: Vector) -> None:
        fold = self.fold_type(
            X, y, n_folds=self.cv, shuffle=self.shuffle, random_state=self.random_state
        )

        for i, (train_indices, test_indices) in enumerate(fold.split):
            X_train, y_train = X[train_indices], y[train_indices]
            X_test, y_test = X[test_indices], y[test_indices]

            self.estimator.fit(X_train, y_train)
            if self.metric:
                train_score = self.estimator.score(X_train, y_train, self.metric)
                test_score = self.estimator.score(X_test, y_test, self.metric)
            else:
                train_score = self.estimator.score(X_train, y_train)
                test_score = self.estimator.score(X_test, y_test)

            self.train_scores_.append(train_score)
            self.test_scores_.append(test_score)

            if self.verbose:
                print(
                    f"[CV] fold {i + 1} -",
                    f"train-score: {train_score:.3f},",
                    f"test-score: {test_score:.3f}",
                )

    def score(self, X: Matrix, y: Vector) -> Tuple[float, float]:
        self._fit(X, y)
        return np.mean(self.train_scores_), np.mean(self.test_scores_)



from typing import Any, Dict

from luma.core.main import Luma
from luma.interface.util import ParamRange


__all__ = (
    "ModelBase",
    "ParadigmBase",
    "MetricBase",
    "VisualBase",
    "NeuralBase",
)


class ModelBase(Luma):
    """
    The ModelBase class serves as a superclass for core machine learning components
    involved in modeling and data transformation. It provides a foundational base
    for classes like Estimator and Transformer, encapsulating shared functionalities
    and properties essential for building, preparing, and processing machine
    learning models. This class is intended to streamline the development of
    machine learning pipelines by offering a unified interface for both model
    creation and data preprocessing tasks.

    Methods
    -------
    Train an estimator or a transformer for further use:
    ```py
        def fit(self, **kwargs) -> Any
    ```

    Set the parameters of an estimator or a transformer:
    ```py
        def set_params(self, **kwargs) -> None
    ```
    This method iterates over the given keyword arguments and sets the
    attributes of the instance accordingly.

    If an attribute corresponding to a given keyword does not exist,
    a message is printed indicating that the attribute was not found.

    Notes
    -----
    Upon the call of `set_params`, the method `check_param_ranges` is
    automatically called after resetting the parameters.

    Inheritances
    ------------
    `Estimator`, `Transformer`, `Optimizer`

    """

    def __validate__(self) -> None:
        return super().__validate__()

    def __alloc__(self, *args, **kwargs) -> None:
        return super().__alloc__(*args, **kwargs)

    def __dealloc__(self) -> None:
        return super().__dealloc__()

    def fit(self, **kwargs) -> Any:
        kwargs

    def set_params(self, ignore_missing: bool = False, **kwargs) -> None:
        for key, val in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, val)
            else:
                if not ignore_missing:
                    print(f"'{type(self).__name__}' has no attribute '{key}'")

        self.check_param_ranges()

    def set_param_ranges(self, range_dict: Dict[str, tuple]) -> None:
        self._param_range_dict = {
            name_: ParamRange(range_, type_)
            for name_, (range_, type_) in range_dict.items()
        }

    def check_param_ranges(self) -> None:
        if not hasattr(self, "_param_range_dict"):
            return
        for name, val in self.__dict__.items():
            if name not in self._param_range_dict:
                continue
            self._param_range_dict[name].check(param_name=name, param_value=val)


class ParadigmBase(Luma):
    """
    The ParadigmBase class is a superclass for different learning paradigms in
    machine learning, such as Supervised and Unsupervised learning. This class
    offers a common base for various types of learning approaches, providing a
    standardized interface and shared properties relevant to different learning
    methodologies. It helps in categorizing and organizing machine learning
    models based on their learning strategy, facilitating the development and
    usage of models tailored to specific types of data and tasks.

    Inheritances
    ------------
    `Supervised`, `Unsupervised`

    """

    def __validate__(self) -> None:
        return super().__validate__()

    def __alloc__(self, *args, **kwargs) -> None:
        return super().__alloc__(*args, **kwargs)

    def __dealloc__(self) -> None:
        return super().__dealloc__()


class MetricBase(Luma):
    """
    The MetricBase class serves as a superclass for evaluation metrics measures
    used in machine learning. It focuses on quantitative measures to assess model
    performance and compute similarities or differences. This class provides a
    unified base for various types of evaluation methods and other metric calculations,
    which are essential for effectively measuring and comparing the efficacy of
    machine learning models and for analyzing data characteristics.

    Inheritances
    ------------
    `Evaluator`, `Distance`

    """

    def __validate__(self) -> None:
        return super().__validate__()

    def __alloc__(self, *args, **kwargs) -> None:
        return super().__alloc__(*args, **kwargs)

    def __dealloc__(self) -> None:
        return super().__dealloc__()


class VisualBase(Luma):
    """
    The VisualBase class is a superclass for visualization tools in machine
    learning. This class offers a common framework for visualizing different
    aspects of machine learning models, datasets, and analysis results. It plays
    a crucial role in simplifying the interpretation and understanding of complex
    machine learning processes, helping users to gain insights into model behavior,
    data patterns, and performance metrics through intuitive visual representations.

    Inheritances
    ------------
    `Visualizer`

    """

    def __validate__(self) -> None:
        return super().__validate__()

    def __alloc__(self, *args, **kwargs) -> None:
        return super().__alloc__(*args, **kwargs)

    def __dealloc__(self) -> None:
        return super().__dealloc__()


class NeuralBase(Luma):
    """
    This class provides the foundational attributes and methods that are common
    across different types of neural network models. It is not intended to be
    instantiated directly but should be subclassed by specific types of neural
    models that implement the specific functionalities.

    Inheritances
    ------------
    `NeuralModel`

    """

    def __validate__(self) -> None:
        return super().__validate__()

    def __alloc__(self, *args, **kwargs) -> None:
        return super().__alloc__(*args, **kwargs)

    def __dealloc__(self) -> None:
        return super().__dealloc__()
